"""Provide the class Image corresponding to an IdxItem.
"""

import logging
import re
from packaging.version import Version
from PySide2 import QtCore, QtGui, QtWidgets
try:
    import vignette
except ImportError:
    vignette = None

log = logging.getLogger(__name__)

# Limit the vignette thumbnailer backends to those dealing with images.
if vignette:
    try:
        # vignette 4.5.2 or newer
        vignette.select_thumbnailer_types([vignette.FILETYPE_IMAGE])
    except AttributeError:
        if hasattr(vignette, 'THUMBNAILER_BACKENDS'):
            # vignette 4.3.0 or newer
            vignette.THUMBNAILER_BACKENDS = [
                vignette.QtBackend(),
                vignette.PilBackend(),
                vignette.MagickBackend()
            ]
        else:
            # vignette is too old to be usable
            vignette = None
    if vignette:
        if not list(vignette.iter_thumbnail_backends()):
            log.warning("Disabling vignette: "
                        "no suitable thumbnailer backend available")
            vignette = None

_thumbs_oriented = (vignette and
                    Version(vignette.__version__) >= Version('5.0.0'))

class ImageNotFoundError(Exception):
    pass


_OrientationTransform = {
    # 1: 'Horizontal (normal)'
    1: ( 1,  0,
         0,  1 ),
    # 2: 'Mirror horizontal'
    2: (-1,  0,
         0,  1 ),
    # 3: 'Rotate 180'
    3: (-1,  0,
         0, -1),
    # 4: 'Mirror vertical'
    4: ( 1,  0,
         0, -1 ),
    # 5: 'Mirror horizontal and rotate 270 CW'
    5: ( 0,  1,
         1,  0 ),
    # 6: 'Rotate 90 CW'
    6: ( 0,  1,
        -1,  0 ),
    # 7: 'Mirror horizontal and rotate 90 CW'
    7: ( 0, -1,
        -1,  0 ),
    # 8: 'Rotate 270 CW'
    8: ( 0, -1,
         1,  0 ),
}
def _get_transform(orientation):
    if orientation is not None:
        return QtGui.QMatrix(*_OrientationTransform[orientation], 0, 0)
    else:
        return QtGui.QMatrix()


class Image(object):

    ThumbnailSize = QtCore.QSize(128, 128)

    def __init__(self, basedir, item):
        self.item = item
        self.fileName = basedir / item.filename
        self.name = item.name or self.fileName.name
        self.init_transform = _get_transform(self.item.orientation)
        self.post_transform = QtGui.QMatrix()

    @property
    def transform(self):
        return self.post_transform * self.init_transform

    def getPixmap(self):
        image = QtGui.QImage(str(self.fileName))
        if image.isNull():
            raise ImageNotFoundError("Cannot load %s." % self.fileName)
        return QtGui.QPixmap.fromImage(image).transformed(self.transform)

    def getThumbPixmap(self):
        if vignette:
            thumbpath = vignette.get_thumbnail(str(self.fileName), 'normal')
            image = QtGui.QImage(thumbpath)
            if image.isNull():
                raise ImageNotFoundError("Cannot load %s." % self.fileName)
            pixmap = QtGui.QPixmap.fromImage(image)
        else:
            image = QtGui.QImage(str(self.fileName))
            if image.isNull():
                raise ImageNotFoundError("Cannot load %s." % self.fileName)
            pixmap = QtGui.QPixmap.fromImage(image)
            pixmap = pixmap.scaled(self.ThumbnailSize, 
                                   QtCore.Qt.KeepAspectRatio, 
                                   QtCore.Qt.SmoothTransformation)
        if _thumbs_oriented:
            transform = self.post_transform
        else:
            transform = self.transform
        pixmap = pixmap.transformed(transform)
        return pixmap

    def rotate(self, a):
        self.post_transform.rotate(a)



"""Use the photoidx.exif module to read EXIF tags from image files.

For the time being, only consider the Orientation tag.  More tests
considering other tags may be added later.
"""

from pathlib import Path
import pytest
import photoidx.exif
from conftest import gettestdata

@pytest.mark.parametrize("o", range(1, 9))
def test_exif_orientation(o):
    testimg = "dsc_1190-o%d.jpg" % o
    filename = Path(gettestdata(testimg))
    exifdata = photoidx.exif.Exif(filename)
    orientation = exifdata.orientation
    assert orientation == o
    assert str(orientation) == photoidx.exif.Orientation.OrientationLabels[o]



#! /usr/bin/python

import sys
import argparse
from PySide2 import QtWidgets
import photoidx.index
import photoidx.idxfilter
from photoidx.qt import ImageViewer


argparser = argparse.ArgumentParser()
argparser.add_argument('-d', '--directory', help="image directory", default=".")
argparser.add_argument('--scale', help="scale factor")
argparser.add_argument('--readOnly', 
                       help='access the index in read only mode', 
                       action='store_const', const=True)
argparser.add_argument('--create', 
                       help='create the index if not present', 
                       action='store_const', const=True)
photoidx.idxfilter.addFilterArguments(argparser)
args = argparser.parse_args()

app = QtWidgets.QApplication([])
try:
    idx = photoidx.index.Index(idxfile=args.directory)
    readOnly = args.readOnly
    dirty = False
except OSError:
    idx = photoidx.index.Index(imgdir=args.directory)
    if args.readOnly:
        readOnly = True
        dirty = False
    else:
        readOnly = not args.create
        dirty = args.create
idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
imageViewer = ImageViewer(idx, idxfilter, args.scale, readOnly, dirty)
sys.exit(app.exec_())



"""pytest configuration.
"""

import os
from pathlib import Path
import shutil
import subprocess
import sys
import tempfile
import pytest
import photoidx


testdir = Path(__file__).parent

def gettestdata(fname):
    fname = testdir / "data" / fname
    assert fname.is_file()
    return str(fname)

@pytest.fixture(scope="module")
def tmpdir(request):
    td = tempfile.mkdtemp(prefix="photoidx-test-")
    yield Path(td)
    shutil.rmtree(td)

def callscript(scriptname, args, stdin=None, stdout=None, stderr=None):
    try:
        script_dir = os.environ['BUILD_SCRIPTS_DIR']
    except KeyError:
        pytest.skip("BUILD_SCRIPTS_DIR is not set.")
    script = Path(script_dir, scriptname)
    cmd = [sys.executable, str(script)] + args
    print("\n>", *cmd)
    subprocess.check_call(cmd, stdin=stdin, stdout=stdout, stderr=stderr)

def pytest_report_header(config):
    """Add information on the package version used in the tests.
    """
    modpath = Path(photoidx.__file__).resolve().parent
    return [ "photoidx: %s" % (photoidx.__version__),
             "          %s" % (modpath) ]



#! /usr/bin/python

import argparse
import photoidx.index
import photoidx.idxfilter
from photoidx.stats import Stats


def create(args):
    idxfile = args.directory if args.update else None
    hashalg = args.checksums.split(',') if args.checksums else []
    with photoidx.index.Index(idxfile=idxfile, imgdir=args.directory,
                              hashalg=hashalg) as idx:
        idx.write()

def ls(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        for i in idxfilter.filter(idx):
            if args.checksum:
                try:
                    checksum = i.checksum[args.checksum]
                except KeyError:
                    continue
                print("%s  %s" % (checksum, i.filename))
            else:
                print(i.filename)

def lstags(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        tags = set()
        for i in idxfilter.filter(idx):
            tags.update(i.tags)
        for t in sorted(tags):
            print(t)

def addtag(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        for i in idxfilter.filter(idx):
            i.tags.add(args.tag)
        idx.write()

def rmtag(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        for i in idxfilter.filter(idx):
            i.tags.discard(args.tag)
        idx.write()

def select(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        for i in idxfilter.filter(idx):
            i.selected = True
        idx.write()

def deselect(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        for i in idxfilter.filter(idx):
            i.selected = False
        idx.write()

def stats(args):
    with photoidx.index.Index(idxfile=args.directory) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter.from_args(args)
        stats = Stats(idxfilter.filter(idx))
        print(str(stats))


argparser = argparse.ArgumentParser()
argparser.add_argument('-d', '--directory', 
                       help="image directory", default=".")
subparsers = argparser.add_subparsers(title='subcommands')

create_parser = subparsers.add_parser('create', help="create the index")
create_parser.add_argument('--checksums', default="md5", 
                           help=("comma separated list of "
                                 "hash algorithms to calculate checksums"))
create_parser.add_argument('--update', action='store_true', 
                           help="add images to an existing index")
create_parser.set_defaults(func=create)

ls_parser = subparsers.add_parser('ls', help="list image files")
ls_parser.add_argument('--checksum', help="hash algorithm to print checksums")
photoidx.idxfilter.addFilterArguments(ls_parser)
ls_parser.set_defaults(func=ls)

lstags_parser = subparsers.add_parser('lstags', help="list tags")
photoidx.idxfilter.addFilterArguments(lstags_parser)
lstags_parser.set_defaults(func=lstags)

addtag_parser = subparsers.add_parser('addtag', help="add tag to images")
addtag_parser.add_argument('tag')
photoidx.idxfilter.addFilterArguments(addtag_parser)
addtag_parser.set_defaults(func=addtag)

rmtag_parser = subparsers.add_parser('rmtag', help="remove tag from images")
rmtag_parser.add_argument('tag')
photoidx.idxfilter.addFilterArguments(rmtag_parser)
rmtag_parser.set_defaults(func=rmtag)

select_parser = subparsers.add_parser('select', 
                                      help="add images to the selection")
photoidx.idxfilter.addFilterArguments(select_parser)
select_parser.set_defaults(func=select)

deselect_parser = subparsers.add_parser('deselect', 
                                        help="remove images from the selection")
photoidx.idxfilter.addFilterArguments(deselect_parser)
deselect_parser.set_defaults(func=deselect)

stats_parser = subparsers.add_parser('stats', help="show statistics")
photoidx.idxfilter.addFilterArguments(stats_parser)
stats_parser.set_defaults(func=stats)

args = argparser.parse_args()
if not hasattr(args, "func"):
    argparser.error("subcommand is required")
args.func(args)



"""Unwanted unicode marker for tags in the index.  (Issue #22)

When tags are set as unicode while the content is pure ASCII, PyYAML
marks them as `!!python/unicode 'tag'`.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]
baseindex = gettestdata("index-create.yaml")
refindex = gettestdata("index-unicode-tags.yaml")

tags = {
    "dsc_4623.jpg": [ "T\u014Dky\u014D", "Ginza" ],
    "dsc_4664.jpg": [ "T\u014Dky\u014D", "Meiji-jing\u016B", "Shint\u014D" ],
    "dsc_4831.jpg": [ "Hakone", "Shint\u014D" ],
    "dsc_5126.jpg": [ "Ky\u014Dto", "Gion" ],
    "dsc_5167.jpg": [ "Ky\u014Dto", "Ry\u014Dan-ji", "Buddha" ],
}


@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    shutil.copy(baseindex, str(tmpdir / ".index.yaml"))
    return tmpdir

def test_tag_unicode(imgdir):
    with photoidx.index.Index(imgdir=imgdir) as idx:
        for item in idx:
            for t in tags[str(item.filename)]:
                item.tags.add(t)
        idx.write()
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"



"""Read and write an index having optional name attributes in it.
"""

import filecmp
import shutil
import pytest
import photoidx.index
from conftest import tmpdir, gettestdata

refindex = gettestdata("index-name.yaml")

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    shutil.copy(refindex, str(tmpdir / ".index.yaml"))
    return tmpdir

def test_readwrite(imgdir):
    """Read the index file and write it out again.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        assert idx[0].name == "ginza.jpg"
        assert idx[1].name is None
        assert idx[3].name == "geisha.jpg"
        idx.write()
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"



"""Call the command line script photo-idx.py.
"""

import datetime
import filecmp
from pathlib import Path
import shutil
import subprocess
import pytest
import yaml
from conftest import tmpdir, gettestdata, callscript

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

refindex = gettestdata("index-create.yaml")

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    return tmpdir


# Note: the default value for the "-d" option to photo-idx is the
# current working directory.  So changing to imgdir before calling
# photo-idx without the "-d" option should be equivalent to calling
# "photo-idx -d imgdir".  We more or less try both variants at random
# in the tests.

@pytest.mark.dependency()
def test_create(imgdir, monkeypatch):
    """Create the index.
    """
    monkeypatch.chdir(str(imgdir))
    callscript("photo-idx.py", ["create"])
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"

@pytest.mark.dependency(depends=["test_create"])
def test_ls_all(imgdir):
    """List all images.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        callscript("photo-idx.py", ["-d", str(imgdir), "ls"], stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == testimgs

@pytest.mark.dependency(depends=["test_create"])
def test_ls_md5(imgdir, monkeypatch):
    """List with --checksum=md5 option.
    """
    md5sum = "/usr/bin/md5sum"
    if not Path(md5sum).is_file():
        pytest.skip("md5sum not found.")
    monkeypatch.chdir(str(imgdir))
    fname = imgdir / "md5"
    with fname.open("wt") as f:
        callscript("photo-idx.py", ["ls", "--checksum=md5"], stdout=f)
    with fname.open("rt") as f:
        cmd = [md5sum, "-c"]
        print(">", *cmd)
        subprocess.check_call(cmd, stdin=f)

@pytest.mark.dependency(depends=["test_create"])
def test_addtag_all(imgdir):
    """Tag all images.
    """
    args = ["-d", str(imgdir), "addtag", "all"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", "all"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == testimgs

@pytest.mark.dependency(depends=["test_create"])
def test_addtag_by_date(imgdir):
    """Select by date.
    """
    args = ["-d", str(imgdir), "addtag", "--date", "2016-03-05", "Hakone"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", "Hakone"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4831.jpg"]

@pytest.mark.dependency(depends=["test_create"])
def test_addtag_by_gpspos(imgdir, monkeypatch):
    """Select by GPS position.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["addtag", "--gpspos", "35.6883 N, 139.7544 E", 
            "--gpsradius", "20.0", "Tokyo"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--tags", "Tokyo"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4623.jpg", "dsc_4664.jpg"]

@pytest.mark.dependency(depends=["test_create"])
def test_addtag_by_files(imgdir, monkeypatch):
    """Select by file names.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["addtag", "Shinto_shrine", "dsc_4664.jpg", "dsc_4831.jpg"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--tags", "Shinto_shrine"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4664.jpg", "dsc_4831.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_all", "test_addtag_by_date", 
             "test_addtag_by_gpspos"]
)
def test_rmtag_by_tag(imgdir, monkeypatch):
    """Remove a tag from images selected by tags.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["rmtag", "--tags", "Tokyo", "all"]
    callscript("photo-idx.py", args)
    args = ["rmtag", "--tags", "Hakone", "all"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--tags", "all"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_5126.jpg", "dsc_5167.jpg"]

@pytest.mark.dependency(depends=["test_create", "test_addtag_all"])
def test_rmtag_all(imgdir, monkeypatch):
    """Remove a tag from all images.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["-d", str(imgdir), "rmtag", "all"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", "all"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == []

@pytest.mark.dependency(depends=["test_create", "test_addtag_by_files"])
def test_ls_by_single_tag(imgdir):
    """Select by one single tag.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", "Shinto_shrine"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4664.jpg", "dsc_4831.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_gpspos", "test_addtag_by_files"]
)
def test_ls_by_mult_tags(imgdir):
    """Select by multiple tags.

    Combining multiple tags acts like an and, it selects only images
    having all the tags set.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", "Tokyo,Shinto_shrine"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4664.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_gpspos", "test_addtag_by_files"]
)
def test_ls_by_neg_tags(imgdir, monkeypatch):
    """Select by negating tags.

    Prepending a tag by an exclamation mark selects the images having
    the tag not set.
    """
    monkeypatch.chdir(str(imgdir))
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--tags", "Tokyo,!Shinto_shrine"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4623.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_date", "test_addtag_by_gpspos", 
             "test_addtag_by_files"]
)
def test_ls_by_empty_tag(imgdir):
    """Select by empty tags.

    The option tags with empty value selects images having no tag.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", "--tags", ""]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_5126.jpg", "dsc_5167.jpg"]

@pytest.mark.dependency(depends=["test_create", "test_addtag_by_gpspos"])
def test_ls_by_date_and_tag(imgdir):
    """Select by date and tags.

    Multiple selection criteria, such as date and tags may be
    combined.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "ls", 
                "--tags", "Tokyo", "--date", "2016-02-28"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4623.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_date", "test_addtag_by_gpspos", 
             "test_addtag_by_files"]
)
def test_lstags_all(imgdir):
    """List tags.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "lstags"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["Hakone", "Shinto_shrine", "Tokyo"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_gpspos", "test_addtag_by_files"]
)
def test_lstags_by_tags(imgdir, monkeypatch):
    """List tags selected by tags.
    """
    monkeypatch.chdir(str(imgdir))
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["lstags", "--tags", "Tokyo"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["Shinto_shrine", "Tokyo"]

@pytest.mark.dependency(depends=["test_create"])
def test_select_by_files(imgdir, monkeypatch):
    """Select by file names.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["select", "dsc_5126.jpg"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--selected"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_5126.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_files", "test_select_by_files"]
)
def test_select_by_tag(imgdir, monkeypatch):
    """Select by tag.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["select", "--tags", "Shinto_shrine"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--selected"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4664.jpg", "dsc_4831.jpg", "dsc_5126.jpg"]

@pytest.mark.dependency(depends=["test_create", "test_select_by_tag"])
def test_deselect_by_files(imgdir, monkeypatch):
    """Deselect by file names.
    """
    monkeypatch.chdir(str(imgdir))
    args = ["deselect", "dsc_4831.jpg"]
    callscript("photo-idx.py", args)
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["ls", "--selected"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        out = f.read().split()
    assert out == ["dsc_4664.jpg", "dsc_5126.jpg"]

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_date", "test_addtag_by_gpspos", 
             "test_addtag_by_files", "test_rmtag_all"]
)
def test_stats_all(imgdir):
    """Show stats.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "stats"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        stats = yaml.safe_load(f)
    assert stats["Count"] == 5
    assert stats["Oldest"] == datetime.datetime(2016, 2, 28, 17, 26, 39)
    assert stats["Newest"] == datetime.datetime(2016, 3, 9, 10, 7, 48)
    assert stats["By date"] == {
        datetime.date(2016, 2, 28) : 1,
        datetime.date(2016, 2, 29) : 1,
        datetime.date(2016, 3, 5) : 1,
        datetime.date(2016, 3, 8) : 1,
        datetime.date(2016, 3, 9) : 1,
    }
    assert stats["By tag"] == {
        "Hakone": 1,
        "Shinto_shrine": 2,
        "Tokyo": 2,
    }

@pytest.mark.dependency(
    depends=["test_create", "test_addtag_by_date", "test_addtag_by_gpspos", 
             "test_addtag_by_files", "test_rmtag_all"]
)
def test_stats_filtered(imgdir):
    """Show stats on a selection.
    """
    fname = imgdir / "out"
    with fname.open("wt") as f:
        args = ["-d", str(imgdir), "stats", "--tags", "Tokyo"]
        callscript("photo-idx.py", args, stdout=f)
    with fname.open("rt") as f:
        stats = yaml.safe_load(f)
    assert stats["Count"] == 2
    assert stats["Oldest"] == datetime.datetime(2016, 2, 28, 17, 26, 39)
    assert stats["Newest"] == datetime.datetime(2016, 2, 29, 11, 37, 51)
    assert stats["By date"] == {
        datetime.date(2016, 2, 28) : 1,
        datetime.date(2016, 2, 29) : 1,
    }
    assert stats["By tag"] == {
        "Shinto_shrine": 1,
        "Tokyo": 2,
    }



"""Get statistics about the items in an index from using class Stats.
"""

import datetime
import shutil
import pytest
import yaml
import photoidx.index
import photoidx.idxfilter
from photoidx.stats import Stats
from conftest import tmpdir, gettestdata


testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    shutil.copy(gettestdata("index-tagged.yaml"), str(tmpdir / ".index.yaml"))
    return tmpdir


def test_stats_all(imgdir):
    """Get statistics on all images.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        stats = Stats(idx)
    assert stats.count == 5
    assert stats.selected == 2
    assert stats.oldest == datetime.datetime(2016, 2, 28, 17, 26, 39)
    assert stats.newest == datetime.datetime(2016, 3, 9, 10, 7, 48)
    assert stats.by_date == {
        datetime.date(2016, 2, 28).toordinal() : 1,
        datetime.date(2016, 2, 29).toordinal() : 1,
        datetime.date(2016, 3, 5).toordinal() : 1,
        datetime.date(2016, 3, 8).toordinal() : 1,
        datetime.date(2016, 3, 9).toordinal() : 1,
    }
    assert stats.by_tag == {
        "Hakone": 1,
        "Shinto_shrine": 2,
        "Tokyo": 2,
    }

def test_stats_all_yaml(imgdir):
    """The string representation of a Stats object is YAML.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        stats = yaml.safe_load(str(Stats(idx)))
    assert stats["Count"] == 5
    assert stats["Selected"] == 2
    assert stats["Oldest"] == datetime.datetime(2016, 2, 28, 17, 26, 39)
    assert stats["Newest"] == datetime.datetime(2016, 3, 9, 10, 7, 48)
    assert stats["By date"] == {
        datetime.date(2016, 2, 28) : 1,
        datetime.date(2016, 2, 29) : 1,
        datetime.date(2016, 3, 5) : 1,
        datetime.date(2016, 3, 8) : 1,
        datetime.date(2016, 3, 9) : 1,
    }
    assert stats["By tag"] == {
        "Hakone": 1,
        "Shinto_shrine": 2,
        "Tokyo": 2,
    }

def test_stats_filtered(imgdir):
    """Get statistics on a selection of images.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags="Tokyo")
        stats = Stats(idxfilter.filter(idx))
    assert stats.count == 2
    assert stats.selected == 1
    assert stats.oldest == datetime.datetime(2016, 2, 28, 17, 26, 39)
    assert stats.newest == datetime.datetime(2016, 2, 29, 11, 37, 51)
    assert stats.by_date == {
        datetime.date(2016, 2, 28).toordinal() : 1,
        datetime.date(2016, 2, 29).toordinal() : 1,
    }
    assert stats.by_tag == {
        "Shinto_shrine": 1,
        "Tokyo": 2,
    }




"""A dialog window to change the filter options.
"""

import datetime
from PySide2 import QtCore, QtWidgets
import photoidx.idxfilter
from photoidx.geo import GeoPosition


class GeoPosEdit(QtWidgets.QLineEdit):
    """A QLineEdit with a suitable size for a GeoPosition.
    """
    def sizeHint(self):
        sh = super().sizeHint()
        fm = self.fontMetrics()
        postext = "\u2014%s\u2014" % GeoPosition("90.0 S, 180.0 E").floatstr()
        sh.setWidth(fm.boundingRect(postext).width())
        return sh


class FilterOption(object):

    def __init__(self, criterion, parent):
        self.groupbox = QtWidgets.QGroupBox("Filter by %s" % criterion)
        self.groupbox.setCheckable(True)
        parent.addWidget(self.groupbox)

    def getOption(self):
        raise NotImplementedError

    def setOption(self, optionValue):
        self.groupbox.setChecked(optionValue is not None)

class TagFilterOption(FilterOption):

    def __init__(self, parent):
        super().__init__("tags", parent)
        self.entry = QtWidgets.QLineEdit()
        label = QtWidgets.QLabel("Tags:")
        label.setBuddy(self.entry)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.entry)
        self.groupbox.setLayout(layout)

    def getOption(self):
        if self.groupbox.isChecked():
            return { 'tags': self.entry.text() }
        else:
            return {}

    def setOption(self, taglist, negtaglist):
        super().setOption(taglist)
        if taglist is not None:
            tags = sorted(taglist)
            negtags = ["!%s" % t for t in sorted(negtaglist)]
            self.entry.setText(",".join(tags + negtags))

class SelectFilterOption(FilterOption):

    def __init__(self, parent):
        super().__init__("selection", parent)
        self.buttonYes = QtWidgets.QRadioButton("selected")
        self.buttonNo = QtWidgets.QRadioButton("not selected")
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.buttonYes)
        layout.addWidget(self.buttonNo)
        self.groupbox.setLayout(layout)

    def getOption(self):
        if self.groupbox.isChecked():
            return { 'select': bool(self.buttonYes.isChecked()) }
        else:
            return {}

    def setOption(self, select):
        super().setOption(select)
        if select is not None:
            if select:
                self.buttonYes.setChecked(QtCore.Qt.Checked)
                self.buttonNo.setChecked(QtCore.Qt.Unchecked)
            else:
                self.buttonYes.setChecked(QtCore.Qt.Unchecked)
                self.buttonNo.setChecked(QtCore.Qt.Checked)

class DateFilterOption(FilterOption):

    def __init__(self, parent):
        super().__init__("date", parent)
        self.startEntry = QtWidgets.QLineEdit()
        startLabel = QtWidgets.QLabel("Start:")
        startLabel.setBuddy(self.startEntry)
        self.endEntry = QtWidgets.QLineEdit()
        endLabel = QtWidgets.QLabel("End:")
        endLabel.setBuddy(self.endEntry)
        layout = QtWidgets.QGridLayout()
        layout.addWidget(startLabel, 0, 0)
        layout.addWidget(self.startEntry, 0, 1)
        layout.addWidget(endLabel, 1, 0)
        layout.addWidget(self.endEntry, 1, 1)
        self.groupbox.setLayout(layout)

    def getOption(self):
        if self.groupbox.isChecked():
            startdate = self.startEntry.text()
            enddate = self.endEntry.text()
            if enddate:
                datestr = "%s--%s" % (startdate, enddate)
            else:
                datestr = startdate
            return { 'date': photoidx.idxfilter.strpdate(datestr) }
        else:
            return {}

    def setOption(self, date):
        super().setOption(date)
        if date is not None:
            self.startEntry.setText(date[0].isoformat())
            self.endEntry.setText(date[1].isoformat())

class GPSFilterOption(FilterOption):

    def __init__(self, parent):
        super().__init__("GPS position", parent)
        self.posEntry = GeoPosEdit()
        posLabel = QtWidgets.QLabel("Position:")
        posLabel.setBuddy(self.posEntry)
        self.radiusEntry = QtWidgets.QLineEdit()
        radiusLabel = QtWidgets.QLabel("Radius:")
        radiusLabel.setBuddy(self.radiusEntry)
        layout = QtWidgets.QGridLayout()
        layout.addWidget(posLabel, 0, 0)
        layout.addWidget(self.posEntry, 0, 1)
        layout.addWidget(radiusLabel, 1, 0)
        layout.addWidget(self.radiusEntry, 1, 1)
        self.groupbox.setLayout(layout)

    def getOption(self):
        if self.groupbox.isChecked():
            return { 'gpspos': GeoPosition(self.posEntry.text()), 
                     'gpsradius': float(self.radiusEntry.text()) }
        else:
            return {}

    def setOption(self, gpspos, gpsradius):
        super().setOption(gpspos)
        if gpspos is not None:
            self.posEntry.setText(gpspos.floatstr())
            self.radiusEntry.setText(str(gpsradius))

class ListFilterOption(FilterOption):

    def __init__(self, parent):
        super().__init__("explicit file names", parent)
        self.entry = QtWidgets.QLineEdit()
        label = QtWidgets.QLabel("Files:")
        label.setBuddy(self.entry)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.entry)
        self.groupbox.setLayout(layout)

    def getOption(self):
        if self.groupbox.isChecked():
            return { 'files': self.entry.text().split() }
        else:
            return {}

    def setOption(self, filelist):
        super().setOption(filelist)
        if filelist is not None:
            self.entry.setText(" ".join(sorted(str(p) for p in filelist)))


class FilterDialog(QtWidgets.QDialog):

    def __init__(self):
        super().__init__()

        mainLayout = QtWidgets.QVBoxLayout()

        self.tagFilterOption = TagFilterOption(mainLayout)
        self.selectFilterOption = SelectFilterOption(mainLayout)
        self.dateFilterOption = DateFilterOption(mainLayout)
        self.gpsFilterOption = GPSFilterOption(mainLayout)
        self.filelistFilterOption = ListFilterOption(mainLayout)

        btn = QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel
        buttonBox = QtWidgets.QDialogButtonBox(btn)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(buttonBox, alignment=QtCore.Qt.AlignHCenter)

        self.setLayout(mainLayout)
        self.setWindowTitle("Filter options")

    def setfilter(self, imgFilter):
        self.imgFilter = imgFilter
        self.tagFilterOption.setOption(imgFilter.taglist, imgFilter.negtaglist)
        self.selectFilterOption.setOption(imgFilter.select)
        self.dateFilterOption.setOption(imgFilter.date)
        self.gpsFilterOption.setOption(imgFilter.gpspos, imgFilter.gpsradius)
        self.filelistFilterOption.setOption(imgFilter.filelist)

    def accept(self):
        filterArgs = {}
        filterArgs.update(self.tagFilterOption.getOption())
        filterArgs.update(self.selectFilterOption.getOption())
        filterArgs.update(self.dateFilterOption.getOption())
        filterArgs.update(self.gpsFilterOption.getOption())
        filterArgs.update(self.filelistFilterOption.getOption())
        self.imgFilter = photoidx.idxfilter.IdxFilter(**filterArgs)
        super().accept()



"""An overview window showing thumbnails of the image set.
"""

import sys
import math
from PySide2 import QtCore, QtWidgets


class ThumbnailWidget(QtWidgets.QLabel):

    def __init__(self, image):
        super().__init__()

        self.setFrameStyle(self.Box | self.Plain)
        self.setLineWidth(4)
        palette = self.palette()
        frameColor = palette.color(self.backgroundRole())
        palette.setColor(self.foregroundRole(), frameColor)
        self.setPalette(palette)

        self.image = image
        self.setImagePixmap()

    def _getOverviewWindow(self):
        w = self
        while not isinstance(w, OverviewWindow):
            w = w.parent()
        return w

    def setImagePixmap(self):
        self.setPixmap(self.image.getThumbPixmap())

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            ovwin = self._getOverviewWindow()
            ovwin.imageViewer.moveCurrentTo(self.image)

    def markActive(self, isActive):
        palette = self.palette()
        if isActive:
            frameColor = QtCore.Qt.blue
        else:
            frameColor = palette.color(self.backgroundRole())
        palette.setColor(self.foregroundRole(), frameColor)
        self.setPalette(palette)


class OverviewWindow(QtWidgets.QMainWindow):

    def __init__(self, imageViewer):
        super().__init__()

        self.imageViewer = imageViewer
        self.numcolumns = 4

        self.setWindowTitle("Overview")
        self.mainLayout = QtWidgets.QGridLayout()
        self._populate()

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(self.mainLayout)
        scrollArea = QtWidgets.QScrollArea()
        scrollArea.setWidget(centralWidget)
        scrollArea.setAlignment(QtCore.Qt.AlignCenter)
        self.setCentralWidget(scrollArea)

        self.closeAct = QtWidgets.QAction("&Close", self)
        self.closeAct.triggered.connect(self.close)

        menu = self.menuBar()

        self.fileMenu = menu.addMenu("&File")
        self.fileMenu.addAction(self.closeAct)

        # Set the width of the window such that the scrollArea just
        # fits.  We need to add 24 to the central widget, 20 for the
        # vertical scroll bar and 4 for the border.
        width = centralWidget.size().width() + 24
        size = self.size()
        size.setWidth(width)
        self.resize(size)

        self.activeWidget = None
        try:
            image = self.imageViewer.selection[self.imageViewer.cur]
            self.markActive(image)
        except IndexError:
            pass

    def _populate(self):
        """Populate the mainLayout with thumbnail images.
        """
        images = self.imageViewer.selection
        ncol = self.numcolumns
        c = 0
        for i in images:
            try:
                thumb = ThumbnailWidget(i)
            except Exception as e:
                print(str(e), file=sys.stderr)
            else:
                self.mainLayout.addWidget(thumb, c // ncol, c % ncol,
                                          QtCore.Qt.AlignCenter)
                c += 1

    def updateThumbs(self):
        """Update the mainLayout with thumbnail images.
        """
        # Note: this code is based on the assumption that no image
        # will ever be added to self.imageViewer.selection and thus we
        # only need to consider removing ThumbnailWidgets, but not to
        # add any.  Furthermore, we assume the order given by
        # self.mainLayout.itemAt() is the the order that the widgets
        # have been added to self.mainLayout and thus the same as
        # self.imageViewer.selection.
        numImages = len(self.imageViewer.selection)
        for i in range(numImages):
            widget = self.mainLayout.itemAt(i).widget()
            image = self.imageViewer.selection[i]
            if widget.image is not image:
                widget.image = image
                widget.setImagePixmap()
        while self.mainLayout.count() > numImages:
            item = self.mainLayout.takeAt(numImages)
            item.widget().deleteLater()

    def getThumbnailWidget(self, image):
        for i in range(self.mainLayout.count()):
            w = self.mainLayout.itemAt(i).widget()
            if w.image is image:
                return w
        else:
            return None

    def markActive(self, image):
        if self.activeWidget:
            self.activeWidget.markActive(False)
        self.activeWidget = self.getThumbnailWidget(image)
        if self.activeWidget:
            self.activeWidget.markActive(True)



"""Provide the class Index which represents an index of photos.
"""

from collections.abc import MutableSequence
import errno
import fcntl
import os
from pathlib import Path
import yaml
from photoidx.idxitem import IdxItem
from photoidx.listtools import LazyList


class AlreadyLockedError(OSError):
    def __init__(self, *args):
        super().__init__(*args)


def _readdir(imgdir, basedir, hashalg, known=set()):
    for f in sorted(imgdir.iterdir()):
        rel = f.relative_to(basedir)
        if f.is_file() and f.suffix == '.jpg' and rel not in known:
            yield IdxItem(filename=rel, basedir=basedir, hashalg=hashalg)


class Index(MutableSequence):

    defIdxFilename = Path(".index.yaml")

    def __init__(self, idxfile=None, imgdir=None, hashalg=['md5']):
        super().__init__()
        self.directory = None
        self.idxfile = None
        self.items = []
        if idxfile:
            self.read(idxfile)
        if imgdir:
            imgdir = Path(imgdir).resolve()
            if not self.directory:
                self.directory = imgdir
            if idxfile:
                self.extend_dir(imgdir, hashalg)
            else:
                newitems = _readdir(imgdir, self.directory, hashalg)
                self.items = LazyList(newitems)

    def extend_dir(self, imgdir, hashalg=['md5']):
        imgdir = Path(imgdir).resolve()
        known = { i.filename for i in self.items }
        newitems = _readdir(imgdir, self.directory, hashalg, known)
        self.items.extend(newitems)

    def close(self):
        if self.idxfile:
            self.idxfile.close()
            self.idxfile = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def __del__(self):
        self.close()

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index):
        return self.items.__getitem__(index)

    def __setitem__(self, index, value):
        self.items.__setitem__(index, value)

    def __delitem__(self, index):
        self.items.__delitem__(index)

    def index(self, value, *args):
        return self.items.index(value, *args)

    def insert(self, index, value):
        self.items.insert(index, value)

    def _get_idxfile(self, fname, flags):
        if fname is not None:
            self.close()
            fname = Path(fname)
            if fname.is_dir():
                fname = fname / self.defIdxFilename
            self.directory = fname.parent.resolve()
            fd = os.open(str(fname), flags, mode=0o666)
            self.idxfile = os.fdopen(fd, "r+t")
        elif self.idxfile:
            self.idxfile.seek(0)
        else:
            if not self.directory:
                self.directory = Path.cwd()
            fname = self.directory / self.defIdxFilename
            fd = os.open(str(fname), flags, mode=0o666)
            self.idxfile = os.fdopen(fd, "r+t")

    def _lockf(self, mode=fcntl.LOCK_SH):
        try:
            fcntl.lockf(self.idxfile.fileno(), mode | fcntl.LOCK_NB)
        except OSError as e:
            if e.errno in (errno.EACCES, errno.EAGAIN):
                e = AlreadyLockedError(*e.args)
            raise e

    def read(self, idxfile=None):
        """Read the index from a file.
        """
        self._get_idxfile(idxfile, os.O_RDWR)
        self._lockf()
        self.items = [ IdxItem(data=i) for i in yaml.safe_load(self.idxfile) ]

    def write(self, idxfile=None):
        """Write the index to a file.
        """
        items = [ i.as_dict() for i in self.items ]
        self._get_idxfile(idxfile, os.O_RDWR|os.O_CREAT)
        self._lockf(mode=fcntl.LOCK_EX)
        yaml.dump(items, self.idxfile, default_flow_style=False)
        self.idxfile.truncate()
        self.idxfile.flush()
        self._lockf()




"""Ordering of tags in the index file.  (Issue #9)

The tags are internally represented as a set, so they are not ordered.
This is correct, as it reflect the nature of tags.  In the index file,
they are represented as a list.  It would be desirable not to have
spurious differences in index files, e.g. two index files should be
equal unless there is a significant difference in the content.  That
is why the tags should have a well defined order in the index file.
"""

import filecmp
import shutil
import pytest
import photoidx.index
import photoidx.idxfilter
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    return tmpdir

# Each test ends up adding the same set of tags to each image.  But
# the order in which the tags are added differs between tests and some
# test also adds spurious tags and removes them later.  The list order
# of elements in a set depends on the history of insertions and
# deletions.  The resulting index file should not be affected by this
# and should always have the same order.  Before fixing Issue #9, the
# arbitrary list order of the set was also in the index file.

tags = {
    "Japan": testimgs,
    "Tokyo": ["dsc_4623.jpg", "dsc_4664.jpg"],
    "Hakone": ["dsc_4831.jpg"],
    "Kyoto": ["dsc_5126.jpg", "dsc_5167.jpg"],
    "Ginza": ["dsc_4623.jpg"],
    "Shinto_shrine": ["dsc_4664.jpg", "dsc_4831.jpg"],
    "Geisha": ["dsc_5126.jpg"],
    "Ryoan-ji": ["dsc_5167.jpg"],
}

@pytest.mark.dependency()
def test_tag_ref(imgdir):
    idxfname = str(imgdir / ".index.yaml")
    reffname = str(imgdir / "index-ref.yaml")
    shutil.copy(gettestdata("index-create.yaml"), idxfname)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        taglist = [ "Japan", "Tokyo", "Hakone", "Kyoto", 
                    "Ginza", "Shinto_shrine", "Geisha", "Ryoan-ji" ]
        for t in taglist:
            idxfilter = photoidx.idxfilter.IdxFilter(files=tags[t])
            for i in idxfilter.filter(idx):
                i.tags.add(t)
        idx.write()
    shutil.copy(idxfname, reffname)

@pytest.mark.dependency(depends=["test_tag_ref"])
def test_tag_shuffle(imgdir):
    """Same as test_tag_ref(), only the order of setting the tags differ.
    """
    idxfname = str(imgdir / ".index.yaml")
    reffname = str(imgdir / "index-ref.yaml")
    shutil.copy(gettestdata("index-create.yaml"), idxfname)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        taglist = [ "Ginza", "Hakone", "Japan", "Geisha", 
                    "Shinto_shrine", "Tokyo", "Kyoto", "Ryoan-ji" ]
        for t in taglist:
            idxfilter = photoidx.idxfilter.IdxFilter(files=tags[t])
            for i in idxfilter.filter(idx):
                i.tags.add(t)
        idx.write()
    assert filecmp.cmp(idxfname, reffname), "index file differs from reference"

@pytest.mark.dependency(depends=["test_tag_ref"])
def test_tag_remove(imgdir):
    """First set all tags on all images, then remove the wrong ones.
    """
    idxfname = str(imgdir / ".index.yaml")
    reffname = str(imgdir / "index-ref.yaml")
    shutil.copy(gettestdata("index-create.yaml"), idxfname)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        taglist = [ "Tokyo", "Shinto_shrine", "Ginza", "Geisha", 
                    "Japan", "Ryoan-ji", "Hakone", "Kyoto" ]
        for t in taglist:
            for i in idx:
                i.tags.add(t)
        for t in taglist:
            for i in idx:
                if str(i.filename) not in tags[t]:
                    i.tags.remove(t)
        idx.write()
    assert filecmp.cmp(idxfname, reffname), "index file differs from reference"

@pytest.mark.dependency(depends=["test_tag_ref"])
def test_tag_extra(imgdir):
    """Add a spurious extra tag first and remove it later.
    """
    idxfname = str(imgdir / ".index.yaml")
    reffname = str(imgdir / "index-ref.yaml")
    shutil.copy(gettestdata("index-create.yaml"), idxfname)
    with photoidx.index.Index(idxfile=imgdir) as idx:
        taglist = [ "Japan", "Tokyo", "Hakone", "Kyoto", 
                    "Ginza", "Shinto_shrine", "Geisha", "Ryoan-ji" ]
        for i in idx:
            i.tags.add("extra")
        for t in taglist:
            idxfilter = photoidx.idxfilter.IdxFilter(files=tags[t])
            for i in idxfilter.filter(idx):
                i.tags.add(t)
        for i in idx:
            i.tags.remove("extra")
        idx.write()
    assert filecmp.cmp(idxfname, reffname), "index file differs from reference"



"""Images in subdirectories.

The Index class is supposed to provide limited support for the images
to be in subdirectories of the directory containing the index file.
All the tooling around does not use this feature for the time being,
though.
"""

import filecmp
import os
from pathlib import Path
import shutil
import subprocess
import pytest
import photoidx.index
import photoidx.idxfilter
from photoidx.geo import GeoPosition
from conftest import tmpdir, gettestdata

testimgs = {
    "Japan": [ "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", "dsc_5167.jpg" ],
    "Quebec": [ "dsc_7157.jpg", "dsc_7490.jpg", "dsc_7582.jpg" ],
}
refindex = gettestdata("index-subdirs.yaml")

hashalg = "md5"
checkprog = "/usr/bin/md5sum"


@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for k in testimgs.keys():
        d = tmpdir / k
        d.mkdir()
        for f in testimgs[k]:
            shutil.copy(gettestdata(f), str(d / f))
    return tmpdir

@pytest.mark.dependency()
def test_create(imgdir):
    """Create the index.
    """
    with photoidx.index.Index(imgdir=imgdir) as idx:
        for k in ("Japan", "Quebec"):
            idx.extend_dir(imgdir / k)
        idx.write()
    idxfile = str(imgdir / ".index.yaml")
    assert filecmp.cmp(refindex, idxfile), "index file differs from reference"

@pytest.mark.dependency(depends=["test_create"])
def test_checksum(imgdir, monkeypatch):
    """Verify the checksum.

    As a side effect, this checks whether the image files are
    accessible by the filename stored in the index.
    """
    if not Path(checkprog).is_file():
        pytest.skip("%s not found." % checkprog)
    fname = imgdir / hashalg
    with photoidx.index.Index(idxfile=imgdir) as idx:
        with fname.open("wt") as f:
            for i in idx:
                print("%s  %s" % (i.checksum[hashalg], i.filename), file=f)
    monkeypatch.chdir(str(imgdir))
    with fname.open("rt") as f:
        cmd = [checkprog, "-c"]
        print(">", *cmd)
        subprocess.check_call(cmd, stdin=f)

@pytest.mark.dependency(depends=["test_create"])
def test_tag(imgdir, monkeypatch):
    """Test tagging of images.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        tokyo = GeoPosition("35.68 N, 139.77 E")
        idxfilter = photoidx.idxfilter.IdxFilter(gpspos=tokyo, gpsradius=500.0)
        for i in idxfilter.filter(idx):
            i.tags.add("Japan")
        quebec = GeoPosition("46.81 N, 71.22 W")
        idxfilter = photoidx.idxfilter.IdxFilter(gpspos=quebec, gpsradius=500.0)
        for i in idxfilter.filter(idx):
            i.tags.add("Quebec")
        idx.write()
    with photoidx.index.Index(idxfile=imgdir) as idx:
        for k in ("Japan", "Quebec"):
            idxfilter = photoidx.idxfilter.IdxFilter(tags=k)
            fnames = [ i.filename for i in idxfilter.filter(idx) ]
            assert fnames == [ Path(k, f) for f in testimgs[k] ]



"""Provide a PySide ImageViewer window.
"""

import sys
from PySide2 import QtCore, QtGui, QtWidgets
import photoidx.index
from photoidx.listtools import LazyList
from photoidx.qt.image import Image
from photoidx.qt.filterDialog import FilterDialog
from photoidx.qt.imageInfoDialog import ImageInfoDialog
from photoidx.qt.overviewWindow import OverviewWindow
from photoidx.qt.tagSelectDialog import TagSelectDialog


class ImageViewer(QtWidgets.QMainWindow):

    def __init__(self, images, imgFilter, 
                 scaleFactor=1.0, readOnly=False, dirty=False):
        super().__init__()

        self.images = images
        self.imgFilter = imgFilter
        self.selection = LazyList(self._filteredImages())
        if readOnly:
            dirty = False
        self.scaleFactor = scaleFactor
        self.readOnly = readOnly
        self.dirty = dirty
        self.cur = 0

        self.imageInfoDialog = ImageInfoDialog(self.images.directory)
        self.overviewwindow = None

        if not self.readOnly:
            taglist = set()
            for i in images:
                taglist |= i.tags
            self.tagSelectDialog = TagSelectDialog(taglist)
        else:
            self.tagSelectDialog = None

        self.filterDialog = FilterDialog()

        self.imageLabel = QtWidgets.QLabel()
        self.imageLabel.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                                      QtWidgets.QSizePolicy.Ignored)
        self.imageLabel.setScaledContents(True)

        self.scrollArea = QtWidgets.QScrollArea()
        self.scrollArea.setBackgroundRole(QtGui.QPalette.Dark)
        self.scrollArea.setWidget(self.imageLabel)
        self.scrollArea.setAlignment(QtCore.Qt.AlignCenter)
        self.setCentralWidget(self.scrollArea)

        maxSize = 0.95 * QtWidgets.QApplication.desktop().screenGeometry().size()
        self.setMaximumSize(maxSize)

        self.saveAct = QtWidgets.QAction("&Save index", self)
        self.saveAct.setShortcut("Ctrl+s")
        self.saveAct.setEnabled(not self.readOnly)
        self.saveAct.triggered.connect(self.saveIndex)
        self.closeAct = QtWidgets.QAction("&Close", self)
        self.closeAct.setShortcut("q")
        self.closeAct.triggered.connect(self.close)
        self.filterOptsAct = QtWidgets.QAction("Filter Options", self)
        self.filterOptsAct.setShortcut("Shift+Ctrl+f")
        self.filterOptsAct.triggered.connect(self.filterOptions)
        self.zoomInAct = QtWidgets.QAction("Zoom &In", self)
        self.zoomInAct.setShortcut(">")
        self.zoomInAct.triggered.connect(self.zoomIn)
        self.zoomOutAct = QtWidgets.QAction("Zoom &Out", self)
        self.zoomOutAct.setShortcut("<")
        self.zoomOutAct.triggered.connect(self.zoomOut)
        self.zoomFitHeightAct = QtWidgets.QAction("Zoom to Fit &Height", self)
        self.zoomFitHeightAct.triggered.connect(self.zoomFitHeight)
        self.zoomFitWidthAct = QtWidgets.QAction("Zoom to Fit &Width", self)
        self.zoomFitWidthAct.triggered.connect(self.zoomFitWidth)
        self.rotateLeftAct = QtWidgets.QAction("Rotate &Left", self)
        self.rotateLeftAct.setShortcut("l")
        self.rotateLeftAct.triggered.connect(self.rotateLeft)
        self.rotateRightAct = QtWidgets.QAction("Rotate &Right", self)
        self.rotateRightAct.setShortcut("r")
        self.rotateRightAct.triggered.connect(self.rotateRight)
        self.fullScreenAct = QtWidgets.QAction("Show &Full Screen", self)
        self.fullScreenAct.setShortcut("f")
        self.fullScreenAct.setCheckable(True)
        self.fullScreenAct.triggered.connect(self.fullScreen)
        self.imageInfoAct = QtWidgets.QAction("Image &Info", self)
        self.imageInfoAct.setShortcut("i")
        self.imageInfoAct.triggered.connect(self.imageInfo)
        self.overviewAct = QtWidgets.QAction("&Overview Window", self)
        self.overviewAct.setShortcut("o")
        self.overviewAct.triggered.connect(self.overview)
        self.prevImageAct = QtWidgets.QAction("&Previous Image", self)
        self.prevImageAct.setShortcut("p")
        self.prevImageAct.setEnabled(False)
        self.prevImageAct.triggered.connect(self.prevImage)
        self.nextImageAct = QtWidgets.QAction("&Next Image", self)
        self.nextImageAct.setShortcut("n")
        self.nextImageAct.setEnabled(self._haveNext())
        self.nextImageAct.triggered.connect(self.nextImage)
        self.selectImageAct = QtWidgets.QAction("&Select Image", self)
        self.selectImageAct.setShortcut("s")
        self.selectImageAct.setEnabled(not self.readOnly)
        self.selectImageAct.triggered.connect(self.selectImage)
        self.deselectImageAct = QtWidgets.QAction("&Deselect Image", self)
        self.deselectImageAct.setShortcut("d")
        self.deselectImageAct.setEnabled(not self.readOnly)
        self.deselectImageAct.triggered.connect(self.deselectImage)
        self.pushForwardAct = QtWidgets.QAction("Push Image &Forward", self)
        self.pushForwardAct.setShortcut("Ctrl+f")
        self.pushForwardAct.setEnabled(not self.readOnly)
        self.pushForwardAct.triggered.connect(self.pushImageForward)
        self.pushBackwardAct = QtWidgets.QAction("Push Image &Backward", self)
        self.pushBackwardAct.setShortcut("Ctrl+b")
        self.pushBackwardAct.setEnabled(not self.readOnly)
        self.pushBackwardAct.triggered.connect(self.pushImageBackward)
        self.tagSelectAct = QtWidgets.QAction("&Tags", self)
        self.tagSelectAct.setShortcut("t")
        self.tagSelectAct.setEnabled(not self.readOnly)
        self.tagSelectAct.triggered.connect(self.tagSelect)

        menu = self.menuBar()

        fileMenu = menu.addMenu("&File")
        fileMenu.addAction(self.saveAct)
        fileMenu.addAction(self.closeAct)
        fileMenu.addAction(self.filterOptsAct)

        viewMenu = menu.addMenu("&View")
        viewMenu.addAction(self.zoomInAct)
        viewMenu.addAction(self.zoomOutAct)
        viewMenu.addAction(self.zoomFitHeightAct)
        viewMenu.addAction(self.zoomFitWidthAct)
        viewMenu.addAction(self.rotateLeftAct)
        viewMenu.addAction(self.rotateRightAct)
        viewMenu.addSeparator()
        viewMenu.addAction(self.fullScreenAct)
        viewMenu.addSeparator()
        viewMenu.addAction(self.imageInfoAct)
        viewMenu.addAction(self.overviewAct)

        imageMenu = menu.addMenu("&Image")
        imageMenu.addAction(self.prevImageAct)
        imageMenu.addAction(self.nextImageAct)
        imageMenu.addAction(self.selectImageAct)
        imageMenu.addAction(self.deselectImageAct)
        imageMenu.addSeparator()
        imageMenu.addAction(self.pushForwardAct)
        imageMenu.addAction(self.pushBackwardAct)
        imageMenu.addSeparator()
        imageMenu.addAction(self.tagSelectAct)

        self.show()
        self._extraSize = self.size() - self.scrollArea.viewport().size()
        self._loadImage()
        self._checkActions()

    def saveIndex(self):
        try:
            self.images.write()
            self.dirty = False
        except photoidx.index.AlreadyLockedError:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setWindowTitle("Index is locked")
            msgBox.setText("Saving the image index failed!")
            msgBox.setInformativeText("Another process is currently "
                                      "accessing the file")
            msgBox.setIcon(QtWidgets.QMessageBox.Critical)
            msgBox.exec_()

    def close(self):
        if self.dirty:
            msgBox = QtWidgets.QMessageBox()
            msgBox.setWindowTitle("Save index?")
            msgBox.setText("The image index been modified.")
            msgBox.setInformativeText("Save changes before closing?")
            msgBox.setIcon(QtWidgets.QMessageBox.Question)
            msgBox.setStandardButtons(QtWidgets.QMessageBox.Save |
                                      QtWidgets.QMessageBox.Discard |
                                      QtWidgets.QMessageBox.Cancel)
            msgBox.setDefaultButton(QtWidgets.QMessageBox.Save)
            ret = msgBox.exec_()
            if ret == QtWidgets.QMessageBox.Save:
                self.saveIndex()
                if self.dirty:
                    return
            elif ret == QtWidgets.QMessageBox.Discard:
                pass
            elif ret == QtWidgets.QMessageBox.Cancel:
                return
        if self.overviewwindow:
            self.overviewwindow.close()
        super().close()

    def _filteredImages(self):
        for item in self.images:
            if self.imgFilter(item):
                yield Image(self.images.directory, item)

    def _haveNext(self):
        """Check whether there is a next image.
        This should be equivalent to self.cur < len(self.selection)-1,
        but without calling len(self.selection).
        """
        try:
            self.selection[self.cur+1]
            return True
        except IndexError:
            return False

    def _setSize(self):
        maxSize = self.maximumSize()
        imgSize = self.imageLabel.pixmap().size()
        if self.scaleFactor is None:
            size = maxSize - self._extraSize
            hscale = size.width() / imgSize.width()
            vscale = size.height() / imgSize.height()
            self.scaleFactor = min(hscale, vscale)
        size = self.scaleFactor * imgSize
        winSize = size + self._extraSize
        if winSize.height() > maxSize.height():
            # Take vertical scrollbar into account.
            sbw = self.scrollArea.verticalScrollBar().size().width()
            winSize += QtCore.QSize(sbw, 0)
        if winSize.width() > maxSize.width():
            # Take horizontal scrollbar into account.
            sbh = self.scrollArea.horizontalScrollBar().size().height()
            winSize += QtCore.QSize(0, sbh)
        self.imageLabel.resize(size)
        if not self.fullScreenAct.isChecked():
            self.resize(winSize)

    def _loadImage(self):
        try:
            image = self.selection[self.cur]
        except IndexError:
            # Nothing to view.
            self.imageLabel.hide()
            if self.overviewwindow:
                self.overviewwindow.markActive(None)
            return
        try:
            pixmap = image.getPixmap()
        except Exception as e:
            print(str(e), file=sys.stderr)
            del self.selection[self.cur]
            if self.overviewwindow:
                self.overviewwindow.updateThumbs()
            self._loadImage()
            return
        self.imageLabel.setPixmap(pixmap)
        self.imageLabel.show()
        self._setSize()
        self.setWindowTitle(image.name)
        if self.overviewwindow:
            self.overviewwindow.markActive(image)

    def _checkActions(self):
        """Enable and disable actions as appropriate.
        """
        self.prevImageAct.setEnabled(self.cur > 0)
        self.pushForwardAct.setEnabled(not self.readOnly and self.cur > 0)
        self.nextImageAct.setEnabled(self._haveNext())
        self.pushBackwardAct.setEnabled(not self.readOnly and self._haveNext())
        try:
            item = self.selection[self.cur].item
        except IndexError:
            # No current image.
            self.imageInfoAct.setEnabled(False)
            self.selectImageAct.setEnabled(False)
            self.deselectImageAct.setEnabled(False)
            self.tagSelectAct.setEnabled(False)
            self.zoomInAct.setEnabled(False)
            self.zoomOutAct.setEnabled(False)
            self.zoomFitHeightAct.setEnabled(False)
            self.zoomFitWidthAct.setEnabled(False)
            self.rotateLeftAct.setEnabled(False)
            self.rotateRightAct.setEnabled(False)
        else:
            self.imageInfoAct.setEnabled(True)
            en_select = not self.readOnly and not item.selected
            en_deselect = not self.readOnly and item.selected
            self.selectImageAct.setEnabled(en_select)
            self.deselectImageAct.setEnabled(en_deselect)
            self.tagSelectAct.setEnabled(not self.readOnly)
            self.zoomInAct.setEnabled(True)
            self.zoomOutAct.setEnabled(True)
            self.zoomFitHeightAct.setEnabled(True)
            self.zoomFitWidthAct.setEnabled(True)
            self.rotateLeftAct.setEnabled(True)
            self.rotateRightAct.setEnabled(True)

    def _reevalFilter(self):
        """Re-evaluate the filter on the current image.
        This is needed if relevant attributes have changed.
        """
        if not self.imgFilter(self.selection[self.cur].item):
            del self.selection[self.cur]
            if self.overviewwindow:
                self.overviewwindow.updateThumbs()
            self._loadImage()
            self._checkActions()

    def get_visible_center(self):
        """Get the center of the current visible area in image coordinates.
        """
        hscroll_pos = self.scrollArea.horizontalScrollBar().value()
        vscroll_pos = self.scrollArea.verticalScrollBar().value()
        scroll_pos = QtCore.QSize(hscroll_pos, vscroll_pos)
        win_center = scroll_pos + self.scrollArea.viewport().size() / 2
        return win_center / self.scaleFactor

    def scroll_visible_center_to(self, pos):
        """Move the scrollbars to center the position in the visible area.
        """
        win_center = self.scaleFactor * pos
        scroll_pos = win_center - self.scrollArea.viewport().size() / 2
        self.scrollArea.horizontalScrollBar().setValue(scroll_pos.width())
        self.scrollArea.verticalScrollBar().setValue(scroll_pos.height())

    def zoomIn(self):
        self.scaleImage(1.6)

    def zoomOut(self):
        self.scaleImage(0.625)

    def zoomFitHeight(self):
        imgHeight = self.imageLabel.pixmap().size().height()
        winHeight = self.scrollArea.viewport().size().height()
        # Leave an internal padding of a few pixel
        winHeight -= 6
        self.scaleFactor = winHeight / imgHeight
        self._setSize()

    def zoomFitWidth(self):
        imgWidth = self.imageLabel.pixmap().size().width()
        winWidth = self.scrollArea.viewport().size().width()
        # Leave an internal padding of a few pixel
        winWidth -= 6
        self.scaleFactor = winWidth / imgWidth
        self._setSize()

    def rotateLeft(self):
        image = self.selection[self.cur]
        image.rotate(-90)
        self.imageLabel.setPixmap(image.getPixmap())
        self._setSize()
        if self.overviewwindow:
            w = self.overviewwindow.getThumbnailWidget(image)
            if w:
                w.setImagePixmap()

    def rotateRight(self):
        image = self.selection[self.cur]
        image.rotate(90)
        self.imageLabel.setPixmap(image.getPixmap())
        self._setSize()
        if self.overviewwindow:
            w = self.overviewwindow.getThumbnailWidget(image)
            if w:
                w.setImagePixmap()

    def fullScreen(self):
        if self.fullScreenAct.isChecked():
            self.showFullScreen()
        else:
            self.showNormal()
            self._setSize()

    def moveCurrentTo(self, newcur):
        if isinstance(newcur, Image):
            newcur = self.selection.index(newcur)
        self.cur = newcur
        self._loadImage()
        self._setSize()
        self._checkActions()

    def prevImage(self):
        if self.cur > 0:
            self.moveCurrentTo(self.cur - 1)

    def nextImage(self):
        if self._haveNext():
            self.moveCurrentTo(self.cur + 1)

    def selectImage(self):
        try:
            self.selection[self.cur].item.selected = True
        except IndexError:
            # No current image.
            pass
        else:
            self.dirty = True
            self.selectImageAct.setEnabled(False)
            self.deselectImageAct.setEnabled(True)
            self._reevalFilter()

    def deselectImage(self):
        try:
            self.selection[self.cur].item.selected = False
        except IndexError:
            # No current image.
            pass
        else:
            self.dirty = True
            self.selectImageAct.setEnabled(True)
            self.deselectImageAct.setEnabled(False)
            self._reevalFilter()

    def imageInfo(self):
        self.imageInfoDialog.setinfo(self.selection[self.cur].item)
        self.imageInfoDialog.exec_()

    def overview(self):
        if not self.overviewwindow:
            self.overviewwindow = OverviewWindow(self)
        self.overviewwindow.show()

    def tagSelect(self):
        item = self.selection[self.cur].item
        self.tagSelectDialog.setCheckedTags(item.tags)
        if self.tagSelectDialog.exec_():
            item.tags = self.tagSelectDialog.checkedTags()
            self.dirty = True
            self._reevalFilter()

    def filterOptions(self):
        self.filterDialog.setfilter(self.imgFilter)
        if self.filterDialog.exec_():
            if self.overviewwindow:
                # The overview window would need to be rebuild in any case.
                self.overviewwindow.close()
                self.overviewwindow = None
            try:
                curidx = self.images.index(self.selection[self.cur].item)
            except IndexError:
                curidx = None
            self.imgFilter = self.filterDialog.imgFilter
            self.selection = LazyList(self._filteredImages())
            if curidx:
                item_i = 0
                for img_i, img in enumerate(self.selection):
                    item_i = self.images.index(img.item, item_i)
                    if item_i >= curidx:
                        cur = img_i
                        break
                else:
                    cur = len(self.selection)
            else:
                cur = 0
            self.moveCurrentTo(cur)

    def scaleImage(self, factor):
        center = self.get_visible_center()
        self.scaleFactor *= factor
        self._setSize()
        self.scroll_visible_center_to(center)

    def pushImageForward(self):
        """Move the current image forward one position in the image order.
        """
        if self.cur > 0:
            i = self.images.index(self.selection[self.cur].item)
            pi = self.images.index(self.selection[self.cur - 1].item)
            item = self.images.pop(i)
            self.images.insert(pi, item)
            img = self.selection.pop(self.cur)
            self.selection.insert(self.cur - 1, img)
            self.dirty = True
            if self.overviewwindow:
                self.overviewwindow.updateThumbs()
            self.moveCurrentTo(self.cur - 1)

    def pushImageBackward(self):
        """Move the current image backward one position in the image order.
        """
        if self._haveNext():
            i = self.images.index(self.selection[self.cur].item)
            ni = self.images.index(self.selection[self.cur + 1].item)
            item = self.images.pop(i)
            self.images.insert(ni - 1, item)
            img = self.selection.pop(self.cur)
            self.selection.insert(self.cur + 1, img)
            self.dirty = True
            if self.overviewwindow:
                self.overviewwindow.updateThumbs()
            self.moveCurrentTo(self.cur + 1)



"""Maintain indices for photo collections

This package maintains indices for photo collections.  The index is
stored as a YAML file and contains metadata and tags describing the
photos.  The photos are accessed read only.
"""

__version__ = "0.10.1"




"""A dialog window to show some informations on the current image.
"""

from PySide2 import QtCore, QtWidgets
from photoidx.exif import Exif


class ImageInfoDialog(QtWidgets.QDialog):

    def __init__(self, basedir):
        super().__init__()
        self.basedir = basedir

        infoLayout = QtWidgets.QGridLayout()
        cameraModelLabel = QtWidgets.QLabel("Camera model:")
        self.cameraModel = QtWidgets.QLabel()
        self.cameraModel.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(cameraModelLabel, 0, 0)
        infoLayout.addWidget(self.cameraModel, 0, 1)
        filenameLabel = QtWidgets.QLabel("File name:")
        self.filename = QtWidgets.QLabel()
        self.filename.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(filenameLabel, 1, 0)
        infoLayout.addWidget(self.filename, 1, 1)
        createDateLabel = QtWidgets.QLabel("Create date:")
        self.createDate = QtWidgets.QLabel()
        self.createDate.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(createDateLabel, 2, 0)
        infoLayout.addWidget(self.createDate, 2, 1)
        orientationLabel = QtWidgets.QLabel("Orientation:")
        self.orientation = QtWidgets.QLabel()
        self.orientation.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(orientationLabel, 3, 0)
        infoLayout.addWidget(self.orientation, 3, 1)
        gpsPositionLabel = QtWidgets.QLabel("GPS position:")
        self.gpsPosition = QtWidgets.QLabel()
        self.gpsPosition.setTextFormat(QtCore.Qt.RichText)
        self.gpsPosition.setOpenExternalLinks(True)
        infoLayout.addWidget(gpsPositionLabel, 4, 0)
        infoLayout.addWidget(self.gpsPosition, 4, 1)
        exposureTimeLabel = QtWidgets.QLabel("Exposure time:")
        self.exposureTime = QtWidgets.QLabel()
        self.exposureTime.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(exposureTimeLabel, 5, 0)
        infoLayout.addWidget(self.exposureTime, 5, 1)
        apertureLabel = QtWidgets.QLabel("F-number:")
        self.aperture = QtWidgets.QLabel()
        self.aperture.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(apertureLabel, 6, 0)
        infoLayout.addWidget(self.aperture, 6, 1)
        isoLabel = QtWidgets.QLabel("ISO speed rating:")
        self.iso = QtWidgets.QLabel()
        self.iso.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(isoLabel, 7, 0)
        infoLayout.addWidget(self.iso, 7, 1)
        focalLengthLabel = QtWidgets.QLabel("Focal length:")
        self.focalLength = QtWidgets.QLabel()
        self.focalLength.setTextFormat(QtCore.Qt.PlainText)
        infoLayout.addWidget(focalLengthLabel, 8, 0)
        infoLayout.addWidget(self.focalLength, 8, 1)

        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
        buttonBox.accepted.connect(self.accept)

        mainLayout = QtWidgets.QVBoxLayout()
        mainLayout.addLayout(infoLayout)
        mainLayout.addWidget(buttonBox, alignment=QtCore.Qt.AlignHCenter)
        self.setLayout(mainLayout)
        self.setWindowTitle("Image Info")

    def setinfo(self, item):
        exifdata = Exif(self.basedir / item.filename)
        self.cameraModel.setText(str(exifdata.cameraModel))
        self.filename.setText(str(item.filename))
        if item.createDate:
            self.createDate.setText(item.createDate.strftime("%a, %x %X"))
        else:
            self.createDate.setText(None)
        if item.orientation:
            self.orientation.setText(str(item.orientation))
        else:
            self.orientation.setText(None)
        pos = item.gpsPosition
        if pos:
            link = "<a href='%s'>%s</a>" % (pos.as_osmurl(), str(pos))
            self.gpsPosition.setText(link)
        else:
            self.gpsPosition.setText(None)
        self.exposureTime.setText(str(exifdata.exposureTime))
        self.aperture.setText(str(exifdata.aperture))
        self.iso.setText(str(exifdata.iso))
        self.focalLength.setText(str(exifdata.focalLength))



"""Maintain indices for photo collections

This package maintains indices for photo collections.  The index is
stored as a YAML file and contains metadata and tags describing the
photos.  The photos are accessed read only.
"""

import setuptools
from setuptools import setup
import setuptools.command.build_py
import distutils.command.sdist
from distutils import log
from glob import glob
from pathlib import Path
import string
try:
    import distutils_pytest
    cmdclass = distutils_pytest.cmdclass
except (ImportError, AttributeError):
    cmdclass = dict()
try:
    import setuptools_scm
    version = setuptools_scm.get_version()
except (ImportError, LookupError):
    try:
        import _meta
        version = _meta.__version__
    except ImportError:
        log.warn("warning: cannot determine version number")
        version = "UNKNOWN"

docstring = __doc__


class meta(setuptools.Command):

    description = "generate meta files"
    user_options = []
    init_template = '''"""%(doc)s"""

__version__ = "%(version)s"
'''
    meta_template = '''
__version__ = "%(version)s"
'''

    def initialize_options(self):
        self.package_dir = None

    def finalize_options(self):
        self.package_dir = {}
        if self.distribution.package_dir:
            for name, path in self.distribution.package_dir.items():
                self.package_dir[name] = convert_path(path)

    def run(self):
        values = {
            'version': self.distribution.get_version(),
            'doc': docstring
        }
        try:
            pkgname = self.distribution.packages[0]
        except IndexError:
            log.warn("warning: no package defined")
        else:
            pkgdir = Path(self.package_dir.get(pkgname, pkgname))
            if not pkgdir.is_dir():
                pkgdir.mkdir()
            with (pkgdir / "__init__.py").open("wt") as f:
                print(self.init_template % values, file=f)
        with Path("_meta.py").open("wt") as f:
            print(self.meta_template % values, file=f)


# Note: Do not use setuptools for making the source distribution,
# rather use the good old distutils instead.
# Rationale: https://rhodesmill.org/brandon/2009/eby-magic/
class sdist(distutils.command.sdist.sdist):
    def run(self):
        self.run_command('meta')
        super().run()
        subst = {
            "version": self.distribution.get_version(),
            "url": self.distribution.get_url(),
            "description": docstring.split("\n")[0],
            "long_description": docstring.split("\n", maxsplit=2)[2].strip(),
        }
        for spec in glob("*.spec"):
            with Path(spec).open('rt') as inf:
                with Path(self.dist_dir, spec).open('wt') as outf:
                    outf.write(string.Template(inf.read()).substitute(subst))


class build_py(setuptools.command.build_py.build_py):
    def run(self):
        self.run_command('meta')
        super().run()


with Path("README.rst").open("rt", encoding="utf8") as f:
    readme = f.read()

setup(
    name = "photoidx",
    version = version,
    description = docstring.split("\n")[0],
    long_description = readme,
    url = "https://github.com/RKrahl/photoidx",
    author = "Rolf Krahl",
    author_email = "rolf@rotkraut.de",
    license = "Apache-2.0",
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    project_urls = dict(
        Source="https://github.com/RKrahl/photoidx",
        Download="https://github.com/RKrahl/photoidx/releases/latest"
    ),
    packages = ["photoidx", "photoidx.qt"],
    scripts = ["scripts/photo-idx.py", "scripts/imageview.py"],
    python_requires = ">=3.6",
    install_requires = ["packaging", "PyYAML", "ExifRead >= 2.2.0"],
    cmdclass = dict(cmdclass, build_py=build_py, sdist=sdist, meta=meta),
)



"""Filter image by various selection criteria.
"""

import datetime
import filecmp
import shutil
import pytest
import photoidx.index
import photoidx.idxfilter
from photoidx.geo import GeoPosition
from conftest import tmpdir, gettestdata

testimgs = [ 
    "dsc_4623.jpg", "dsc_4664.jpg", "dsc_4831.jpg", 
    "dsc_5126.jpg", "dsc_5167.jpg" 
]
testimgfiles = [ gettestdata(i) for i in testimgs ]

@pytest.fixture(scope="module")
def imgdir(tmpdir):
    for fname in testimgfiles:
        shutil.copy(fname, str(tmpdir))
    shutil.copy(gettestdata("index-tagged.yaml"), str(tmpdir / ".index.yaml"))
    return tmpdir

def test_by_date(imgdir):
    """Select by date.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        date = (datetime.datetime(2016, 3, 5), datetime.datetime(2016, 3, 6))
        idxfilter = photoidx.idxfilter.IdxFilter(date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4831.jpg"]

def test_by_gpspos(imgdir):
    """Select by GPS position.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        pos = GeoPosition("35.6883 N, 139.7544 E")
        idxfilter = photoidx.idxfilter.IdxFilter(gpspos=pos, gpsradius=20.0)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4623.jpg", "dsc_4664.jpg"]

def test_by_files(imgdir):
    """Select by file names.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        files = ["dsc_4664.jpg", "dsc_4831.jpg"]
        idxfilter = photoidx.idxfilter.IdxFilter(files=files)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == files

def test_by_single_tag(imgdir):
    """Select by one single tag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags="Shinto_shrine")
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4664.jpg", "dsc_4831.jpg"]

def test_by_mult_tags(imgdir):
    """Select by multiple tags.

    Combining multiple tags acts like an and, it selects only images
    having all the tags set.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags="Tokyo,Shinto_shrine")
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4664.jpg"]

def test_by_neg_tags(imgdir):
    """Select by negating tags.

    Prepending a tag by an exclamation mark selects the images having
    the tag not set.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags="Tokyo,!Shinto_shrine")
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4623.jpg"]

def test_by_empty_tag(imgdir):
    """Select by empty tags.

    The empty string as tag selects images having no tag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(tags="")
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_5126.jpg", "dsc_5167.jpg"]

def test_by_date_and_tag(imgdir):
    """Select by date and tags.

    Multiple selection criteria, such as date and tags may be
    combined.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        date = (datetime.datetime(2016, 2, 28), datetime.datetime(2016, 2, 29))
        idxfilter = photoidx.idxfilter.IdxFilter(tags="Tokyo", date=date)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4623.jpg"]

def test_by_selected(imgdir):
    """Select by selected flag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(select=True)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4664.jpg", "dsc_5126.jpg"]

def test_by_selected_and_tag(imgdir):
    """Select by selected flag and tag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(select=True, tags="Tokyo")
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4664.jpg"]

def test_by_not_selected(imgdir):
    """Select by not-selected flag.
    """
    with photoidx.index.Index(idxfile=imgdir) as idx:
        idxfilter = photoidx.idxfilter.IdxFilter(select=False)
        fnames = [ str(i.filename) for i in idxfilter.filter(idx) ]
        assert fnames == ["dsc_4623.jpg", "dsc_4831.jpg", "dsc_5167.jpg"]




"""Define filter for photo indices by command line arguments.
"""

import argparse
import datetime
from pathlib import Path
import re
from photoidx.geo import GeoPosition


_datere = re.compile(r'''^
    (?P<y0>\d{4})-(?P<m0>\d{2})-(?P<d0>\d{2})              # start date
    (?:[T ](?P<h0>\d{2}):(?P<min0>\d{2}):(?P<s0>\d{2}))?   # optional start time
    (?:\s*(?:--|/)\s*                                      # delimeter
      (?P<y1>\d{4})-(?P<m1>\d{2})-(?P<d1>\d{2})            # end date
      (?:[T ](?P<h1>\d{2}):(?P<min1>\d{2}):(?P<s1>\d{2}))? # optional end time
    )?$''', re.X)

def strpdate(s):
    """Parse a date and time intervall.
    """
    def _dt(args):
        year, month, day, hour, minute, second = args
        if hour is None:
            return datetime.datetime(int(year), int(month), int(day))
        else:
            return datetime.datetime(int(year), int(month), int(day), 
                                     int(hour), int(minute), int(second))
    match = re.match(_datere, s)
    if not match:
        raise ValueError("Invalid date value '%s'" % s)
    startdate = _dt(match.group('y0', 'm0', 'd0', 'h0', 'min0', 's0'))
    if match.group('d1') is None:
        if match.group('h0') is None:
            enddate = startdate + datetime.timedelta(days=1)
        else:
            enddate = startdate + datetime.timedelta(seconds=1)
    else:
        enddate = _dt(match.group('y1', 'm1', 'd1', 'h1', 'min1', 's1'))
    return (startdate, enddate)


class IdxFilter(object):

    @classmethod
    def from_args(cls, args):
        kwargs = {}
        for a in ("tags", "select", "date", "gpspos", "gpsradius", "files"):
            kwargs[a] = getattr(args, a)
        return cls(**kwargs)

    def __init__(self, 
                 tags=None, select=None, date=None, 
                 gpspos=None, gpsradius=3.0, files=None):
        if tags is not None:
            self.taglist = set()
            self.negtaglist = set()
            if tags:
                for t in tags.split(","):
                    if t.startswith("!"):
                        self.negtaglist.add(t[1:])
                    else:
                        self.taglist.add(t)
        else:
            self.taglist = None
            self.negtaglist = None
        self.select = select
        self.date = date
        if gpspos:
            self.gpspos = gpspos
            self.gpsradius = gpsradius
        else:
            self.gpspos = None
            self.gpsradius = None
        if files:
            self.filelist = { Path(f) for f in files }
        else:
            self.filelist = None

    def __call__(self, item):
        if self.filelist and not item.filename in self.filelist:
            return False
        if self.taglist is not None:
            if not self.taglist <= item.tags:
                return False
            if self.negtaglist & item.tags:
                return False
            if not self.taglist and not self.negtaglist and item.tags:
                return False
        if self.select is not None and item.selected != self.select:
            return False
        if self.date and (item.createDate < self.date[0] or 
                          item.createDate >= self.date[1]):
            return False
        if self.gpspos:
            if (not item.gpsPosition or
                (item.gpsPosition - self.gpspos) > self.gpsradius):
                return False
        return True

    def filter(self, idx):
        return filter(self, idx)


def addFilterArguments(argparser):
    def _strpdate_arg(s):
        try:
            return strpdate(s)
        except ValueError:
            raise argparse.ArgumentTypeError("Invalid date value '%s'" % s)

    argparser.add_argument('--tags', 
                           help="select images by comma separated list of tags")
    argparser.add_argument('--selected', 
                           help='select images in the selection', 
                           dest='select', action='store_const', const=True)
    argparser.add_argument('--not-selected', 
                           help='select images not in the selection', 
                           dest='select', action='store_const', const=False)
    argparser.add_argument('--date', type=_strpdate_arg, 
                           help="select images by date")
    argparser.add_argument('--gpspos', 
                           help="select images by GPS position", 
                           type=GeoPosition)
    argparser.add_argument('--gpsradius', type=float, default=3.0, 
                           help="radius around GPS position in km", 
                           metavar='DISTANCE')
    argparser.add_argument('files', nargs='*', 
                           help="select images by file name")
    return argparser
